--------------------------------------------------------------------------------
-- From Structured1 to Deriving1, and Structured0 to Deriving0
--------------------------------------------------------------------------------

{-# LANGUAGE GADTs                  #-}
{-# LANGUAGE TypeOperators          #-}
{-# LANGUAGE TypeFamilies           #-}
{-# LANGUAGE FlexibleContexts       #-}
{-# LANGUAGE FlexibleInstances      #-}
{-# LANGUAGE PolyKinds              #-}
{-# LANGUAGE DataKinds              #-}
{-# LANGUAGE MultiParamTypeClasses  #-}
{-# LANGUAGE ScopedTypeVariables    #-}
{-# LANGUAGE UndecidableInstances   #-}


module Conversion.S2D1 where

import qualified Library.Structured as S
import qualified Library.Deriving   as D

--------------------------------------------------------------------------------
-- Conversion of representations (types)
--------------------------------------------------------------------------------

type family S2D (r :: k) :: D.Un * -- overly polymorphic, for overloading
type instance S2D ('S.UData md cs) = S2D cs

-- Constructors
type instance S2D (S.Empty               :: S.Tree (S.UCon k)) = D.VV
type instance S2D (S.Leaf ('S.UCon mc x) :: S.Tree (S.UCon k)) = D.MM mc (S2D x)
type instance S2D (S.Bin x y             :: S.Tree (S.UCon k)) = (D.:+:) (S2D x) (S2D y)

-- Arguments to constructors
type instance S2D (S.Empty               :: S.Tree (S.USel k)) = D.UU
type instance S2D (S.Leaf ('S.USel ms x) :: S.Tree (S.USel k)) = S2D x
type instance S2D (S.Bin x y             :: S.Tree (S.USel k)) = (D.:**:) (S2D x) (S2D y)

-- Fields
type instance S2D (S.UK1 i a)    = D.KK i a
type instance S2D (S.URec i a)   = D.REC i a
type instance S2D (S.UPar)       = D.PAR
type instance S2D ((S.:.:) t f)  = (D.:.:) t (S2D f)

--------------------------------------------------------------------------------
-- Value-level conversion
--------------------------------------------------------------------------------

class Convert (r :: k) where
  s2d :: S.In r p -> D.In (S2D r) p
  d2s :: D.In (S2D r) p -> S.In r p

instance (Convert cs) => Convert ('S.UData md cs) where
  s2d (S.D1 x) = s2d x
  d2s x = S.D1 (d2s x)

-- Constructors: none
instance Convert (S.Empty :: S.Tree (S.UCon k)) where
  s2d _ = error "no element of type void"
  d2s _ = error "no element of type void"

-- Constructors: picking one
instance (Convert x, Convert y)
        => Convert (S.Bin x y :: S.Tree (S.UCon k)) where
  s2d (S.L1 x) = D.L1 (s2d x)
  s2d (S.R1 x) = D.R1 (s2d x)
  d2s (D.L1 x) = S.L1 (d2s x)
  d2s (D.R1 x) = S.R1 (d2s x)

-- Constructors: metadata, proceed to constructor arguments
instance (Convert x)
        => Convert (S.Leaf ('S.UCon mc x) :: S.Tree (S.UCon k)) where
  s2d (S.C1 x) = D.M1 (s2d x)
  d2s (D.M1 x) = S.C1 (d2s x)

-- Constructor arguments: none
instance Convert (S.Empty :: S.Tree (S.USel k)) where
  s2d S.U1 = D.U1
  d2s D.U1 = S.U1

-- Constructor arguments: tupling
instance (Convert l, Convert r)
        => Convert (S.Bin l r :: S.Tree (S.USel k)) where
  s2d ((S.:*:) x y) = (D.:*:) (s2d x) (s2d y)
  d2s ((D.:*:) x y) = (S.:*:) (d2s x) (d2s y)

-- Constructor arguments: metadata, proceed to actual argument
instance (Convert x)
        => Convert (S.Leaf ('S.USel ms x) :: S.Tree (S.USel k)) where
  s2d (S.S1 x) = s2d x
  d2s x        = S.S1 (d2s x)

-- Actual arguments
instance Convert (S.UK1 i a :: S.UArg *) where
  s2d (S.K1 x) = D.K1 x
  d2s (D.K1 x) = S.K1 x

instance Convert (S.URec i a :: S.UArg *) where
  s2d (S.Rec x) = D.Rec1 x
  d2s (D.Rec1 x) = S.Rec x

instance Convert (S.UPar :: S.UArg *) where
  s2d (S.Par p) = D.Par1 p
  d2s (D.Par1 p) = S.Par p

instance (D.GFunctor t, Convert f)
        => Convert ((S.:.:) t f :: S.UArg *) where
  s2d (S.Comp p) = D.Comp1 (D.gmap s2d p) -- gmap could come from somewhere else
  d2s (D.Comp1 p) = S.Comp (D.gmap d2s p)

--------------------------------------------------------------------------------
-- Embedding conversion
--------------------------------------------------------------------------------

-- There is no general instance; each one has to be given individually. See
-- Test.S2D1.
